import albumentations as A
import cv2
import numpy as np
import numpy.typing as npt

from albumentations.pytorch import ToTensorV2
from albumentations.core.transforms_interface import ImageOnlyTransform


def resize_fixed_height(img: npt.NDArray[np.uint8], new_height: int = 105):
    # From the paper: height is fixed to 105 pixels, width is scaled
    # to keep aspect ratio.
    height, width = img.shape[:2]
    new_width = round(new_height * width / height)
    return cv2.resize(img, (new_width, new_height), interpolation=cv2.INTER_LANCZOS4)


def split_patches_np(img: npt.NDArray[np.uint8], step: int, drop_last: bool):
    height, width = img.shape

    patches = []
    for x in range(0, width, step):
        patches.append(img[0:height, x : x + step])

    # Fixup the last patch instead of dropping it because
    # its width is smaller than 105. When cropping and
    # the patch is smaller than the needed area, it gets filled
    # with black pixels. We should recolor them instead of discarding.
    available_width = width % step
    if available_width != 0:
        if drop_last:
            patches.pop()
        else:
            patches[-1] = np.append(
                patches[-1],
                np.full((height, step - available_width), 255, dtype="uint8"),
                axis=1,
            )

    return patches


class PickRandomPatch(ImageOnlyTransform):
    """
    Pick a random 105x105 patch.
    """

    def __init__(
        self,
        constrained_patches=True,
        drop_last=False,
        always_apply=False,
        p=1.0,
    ) -> None:
        super(PickRandomPatch, self).__init__(always_apply, p)
        self.constrained_patches = constrained_patches
        self.drop_last = drop_last

    def apply(self, img, **params):
        # If patches are constrained, then we split the
        # image into 105x105 boxes and pick one of the boxes.
        # Otherwise, we pick a random start coordinate and
        # build a box from there.

        _, width = img.shape
        if width <= 105:
            return np.append(
                img,
                np.full((105, 105 - width), 255, dtype="uint8"),
                axis=1,
            )

        if not self.constrained_patches:
            start_x = np.random.randint(0, width - 105)
            return img[0:105, start_x : start_x + 105]

        patches = split_patches_np(img, 105, self.drop_last)
        patch_index = np.random.randint(0, len(patches))
        return patches[patch_index]


class VariableAspectRatio(ImageOnlyTransform):
    """
    The image, with heigh fixed, is squeezed in width by a random
    ratio, drawn from a uniform distribution between a range.
    """

    def __init__(self, ratio_range, always_apply=False, p=1.0) -> None:
        super(VariableAspectRatio, self).__init__(always_apply, p)
        self.ratio_range = ratio_range

    def apply(self, img, **params):
        height, width = img.shape
        ratio = np.random.uniform(low=self.ratio_range[0], high=self.ratio_range[1])
        new_width = round(width * ratio)
        squeezed = cv2.resize(img, (new_width, height), cv2.INTER_LANCZOS4)
        return np.array(squeezed)

    def get_transform_init_args_names(self):
        return ("ratio_range",)


class Squeezing(ImageOnlyTransform):
    """
    Add a "squeezing" operation = "we introduce a squeezing operation, that
    scales the width of the height-normalized image to be of a constant ratio relative
    to the height (2.5 in all our experiments). Note that the squeezing operation is
    equivalent to producing long rectangular input patches."
    """

    def __init__(self, squeeze_ratio, always_apply=False, p=1.0) -> None:
        super(Squeezing, self).__init__(always_apply, p)
        self.squeeze_ratio = squeeze_ratio

    def apply(self, img, **params):
        height, width = img.shape
        new_width = round(height * self.squeeze_ratio)
        squeezed = cv2.resize(img, (new_width, height), cv2.INTER_LANCZOS4)
        return np.array(squeezed, dtype="uint8")

    def get_transform_init_args_names(self):
        return ("squeeze_ratio",)


class ResizeHeight(ImageOnlyTransform):
    """
    Resize the image height keeping the aspect ratio.
    """

    def __init__(self, target_height: int, always_apply=False, p=1.0) -> None:
        super(ResizeHeight, self).__init__(always_apply, p)
        self.target_height = target_height

    def apply(self, img, **params):
        resized = resize_fixed_height(img, self.target_height)
        return np.array(resized)


def get_deepfont_base_augmentations() -> A.Compose:
    return A.Compose(
        [
            # Parameters are taken from the paper.
            A.GaussNoise(var_limit=(3.0, 4.0), p=0.5),
            A.GaussianBlur(blur_limit=(1, 3), p=0.5),
            # Images are expected to be grayscale and have a white background, so
            # use the 255 value as a filler.
            A.Affine(rotate=[-3, 3], cval=255, p=0.5),
            A.MultiplicativeNoise(multiplier=(0.3, 0.8), p=0.5),
        ]
    )


def get_deepfont_feature_enhancement() -> A.Compose:
    return A.Sequential(
        [
            # From the deepfont paper, between 5/6 and 7/6.
            VariableAspectRatio(ratio_range=[0.83, 1.17], always_apply=True),
            Squeezing(squeeze_ratio=2.5, always_apply=True),
        ]
    )


def get_random_square_patch() -> A.Compose:
    return A.Sequential(
        [
            PickRandomPatch(constrained_patches=False, always_apply=True),
            A.ToFloat(255, always_apply=True),
            ToTensorV2(),
        ]
    )


def get_deepfont_full_augmentations() -> A.Compose:
    return A.Compose(
        [
            A.Sequential(
                [
                    ResizeHeight(target_height=105, always_apply=True),
                    get_deepfont_base_augmentations(),
                    get_deepfont_feature_enhancement(),
                    PickRandomPatch(constrained_patches=False, always_apply=True),
                    A.ToFloat(255, always_apply=True),
                    ToTensorV2(),
                ]
            ),
        ]
    )


def get_random_square_patch_augmentation() -> A.Compose:
    return A.Compose(
        [
            ResizeHeight(target_height=105, always_apply=True),
            # Testing still requires VAR and squeezing.
            get_deepfont_feature_enhancement(),
            get_random_square_patch(),
        ]
    )


def get_test_augmentations(squeeze_ratio: float) -> A.Compose:
    return A.Sequential(
        [
            ResizeHeight(target_height=105, always_apply=True),
            Squeezing(squeeze_ratio=squeeze_ratio, always_apply=True),
        ]
    )



import argparse
import pytorch_lightning as L
from pytorch_lightning.callbacks import (
    EarlyStopping,
    LearningRateMonitor,
    ModelCheckpoint,
    ModelSummary,
)
import torch

from fontina.adobevfr_dataset import AdobeVFRDataset
from fontina.augmentation_utils import (
    get_deepfont_full_augmentations,
    get_random_square_patch_augmentation,
)
from fontina.augmented_dataset import AugmentedDataset

from fontina.config import load_config
from fontina.models.deepfont import DeepFont
from fontina.models.lightning_generate_callback import GenerateCallback
from fontina.models.lightning_wrappers import DeepFontAutoencoderWrapper, DeepFontWrapper

from torch.utils.data import DataLoader
from torchvision import datasets, transforms

# The torchvision `DataLoader` uses Pillow under the hood to
# load image files. However PIL will fail to load some PNG
# files by throwing a zlib decompression error:
# "ValueError: Decompressed Data Too Large". Setting
# `LOAD_TRUNCATED_IMAGES = True` mitigates this problem.
from PIL import Image, ImageFile, UnidentifiedImageError

ImageFile.LOAD_TRUNCATED_IMAGES = True


def get_parser():
    parser = argparse.ArgumentParser(description="Fontina training")
    parser.add_argument(
        "-c",
        "--config",
        metavar="CONFIG-FILE",
        help="path to the configuration file",
        required=True,
    )
    parser.add_argument(
        "-r",
        "--resume-from",
        metavar="CHECKPOINT-FILE",
        help="path to the checkpoint file",
        required=False,
    )
    return parser


def load_and_split_data(train_config):
    all_data = datasets.ImageFolder(
        root=train_config["data_root"],
        # Important: albumentation can't set grayscale and output only one
        # channel, so do it here.
        transform=transforms.Grayscale(num_output_channels=1),
        target_transform=None,
    )
    full_data_num_classes = len(all_data.classes)

    split_ratios = (
        [
            train_config["train_ratio"],
            train_config["validation_ratio"],
            train_config["test_ratio"],
        ]
        if train_config["run_test_cycle"]
        else [
            train_config["train_ratio"],
            1.0 - train_config["train_ratio"],
        ]
    )
    splits = torch.utils.data.random_split(all_data, split_ratios)

    train_set_processed = AugmentedDataset(
        splits[0], full_data_num_classes, get_deepfont_full_augmentations()
    )
    validation_set_processed = AugmentedDataset(
        splits[1], full_data_num_classes, get_random_square_patch_augmentation()
    )
    test_set_processed = (
        AugmentedDataset(
            splits[2], full_data_num_classes, get_random_square_patch_augmentation()
        )
        if train_config["run_test_cycle"]
        else None
    )

    return (
        full_data_num_classes,
        train_set_processed,
        validation_set_processed,
        test_set_processed,
    )


def load_adobevfr_dataset(train_config):
    all_train_data = AdobeVFRDataset(
        f"{train_config['data_root']}/BCF format/VFR_syn_train",
        "train",
        get_deepfont_full_augmentations(),
    )
    num_labels = all_train_data.num_labels

    # From the DeepFont paper: "We first train the SCAE on both synthetic and
    # real-world data in a unsupervised way [...]". When training the autoencoder,
    # we merge the real and synthetic datasets for training purposes.
    if train_config["only_autoencoder"]:

        def is_valid_image(path):
            try:
                _ = Image.open(path)
                return True
            except (UnidentifiedImageError, ValueError):
                print(f"Failed to load image: {path}")
                return False

        real_data = datasets.ImageFolder(
            root=f"{train_config['data_root']}/Raw Image/VFR_real_u",
            # Important: albumentation can't set grayscale and output only one
            # channel, so do it here.
            transform=transforms.Grayscale(num_output_channels=1),
            target_transform=None,
            is_valid_file=is_valid_image,
        )
        real_data_processed = AugmentedDataset(
            real_data, 1, get_deepfont_full_augmentations()
        )

        # Override 'all_train_data' with the joined dataset.
        all_train_data = torch.utils.data.ConcatDataset(
            [all_train_data, real_data_processed]
        )

    # Although the AdobeVFR dataset readme says that VFR_syn_val contains
    # the validation for the same classes as VFR_syn_train, that doens't
    # seem to be the case: the former contains 2383 classes, the latter
    # 4383. Instead of using it, let's split the rain set.
    splits = torch.utils.data.random_split(all_train_data, [0.95, 0.05])

    train_set_processed = AugmentedDataset(splits[0], num_labels, None)
    validation_set_processed = AugmentedDataset(splits[1], num_labels, None)

    """
    validation_set_processed = AdobeVFRDataset(
        f"{train_config['data_root']}/BCF format/VFR_syn_val",
        "val",
        get_random_square_patch_augmentation(),
    )
    """
    test_set_processed = (
        AdobeVFRDataset(
            f"{train_config['data_root']}/BCF format/VFR_real_test",
            "vfr_large",
            get_random_square_patch_augmentation(),
        )
        if train_config["run_test_cycle"]
        else None
    )

    return (
        num_labels,
        train_set_processed,
        validation_set_processed,
        test_set_processed,
    )


def main():
    args = get_parser().parse_args()

    config = load_config(args.config)

    train_config = config["training"]

    if "fixed_seed" in train_config:
        L.seed_everything(train_config["fixed_seed"])

    (
        full_data_num_classes,
        train_set_processed,
        validation_set_processed,
        test_set_processed,
    ) = (
        load_and_split_data(train_config)
        if train_config["dataset_type"] == "raw-images"
        else load_adobevfr_dataset(train_config)
    )

    batch_size = train_config["batch_size"]
    num_workers = max(train_config["num_workers"] // 3, 1)

    train_loader = DataLoader(
        train_set_processed,
        batch_size=batch_size,
        num_workers=num_workers,
        shuffle=True,
        persistent_workers=True,
        pin_memory=True,
    )
    validation_loader = DataLoader(
        validation_set_processed,
        batch_size=batch_size,
        num_workers=num_workers,
        shuffle=False,
    )

    # Try to make training a bit faster without sacrificing precision (as suggested
    # by lightning AI).
    torch.set_float32_matmul_precision("high")

    train_autoencoder = train_config["only_autoencoder"]
    model = None
    if train_autoencoder:
        model = DeepFontAutoencoderWrapper()
    else:
        autoenc_wrapper = DeepFontAutoencoderWrapper.load_from_checkpoint(
            train_config["scae_checkpoint_file"]
        )

        model = DeepFontWrapper(
            model=DeepFont(
                autoencoder=autoenc_wrapper.autoencoder,
                num_classes=full_data_num_classes,
            ),
            num_classes=full_data_num_classes,
            learning_rate=train_config["learning_rate"],
        )

    # Define the set of callbacks to use during the training process.
    training_callbacks = [
        EarlyStopping(monitor="val_loss", mode="min"),
        LearningRateMonitor(logging_interval="step"),
        ModelCheckpoint(
            dirpath=train_config["output_dir"],
            save_top_k=2,
            monitor="val_loss",
            filename="deepfont-{epoch:02d}-{val_loss:.4f}-{val_accuracy:.4f}",
            save_last=True,
        ),
        ModelSummary(-1),
    ]

    if train_autoencoder:
        # When training the autoencoder,
        sample_index = 10 % len(train_set_processed)
        training_callbacks.append(
            GenerateCallback(
                train_set_processed[sample_index][0].unsqueeze(0), every_n_epochs=5
            ),
        )

    trainer = L.Trainer(
        # If ever doing gradient clipping here, remember that this
        # has an effect on training: it will take likely twice as
        # long to get to the same accuracy!
        # gradient_clip_algorithm="norm",
        # gradient_clip_val=1.0,
        max_epochs=train_config["epochs"],
        callbacks=training_callbacks,
    )

    trainer.fit(
        model=model,
        train_dataloaders=train_loader,
        val_dataloaders=validation_loader,
        ckpt_path=args.resume_from if args.resume_from else None,
    )

    # Test the model
    if "run_test_cycle" in train_config and train_config["run_test_cycle"]:
        test_loader = DataLoader(
            test_set_processed,
            batch_size=batch_size,
            num_workers=num_workers,
            shuffle=False,
            persistent_workers=True,
        )
        trainer.test(model, test_loader)


if __name__ == "__main__":
    main()



import argparse
import random
import rstr

from pathlib import Path, PurePath
from typing import Optional

from trdg.generators import GeneratorFromRandom, GeneratorFromStrings

from fontina.config import load_config


class GeneratorFromRegex(GeneratorFromStrings):
    def __init__(
        self, regex_template: str, count: int, random_character_spacing: bool, **kwargs
    ):
        super().__init__(
            count=count,
            strings=self._random_strings(
                count,
                regex_template,
            ),
            **kwargs,
        )
        self.random_character_spacing = random_character_spacing

    def _random_strings(self, num_to_generate: int, regex_template: str):
        return [rstr.xeger(regex_template) for _ in range(0, num_to_generate)]

    def _clamp(self, x, range_min, range_max):
        return max(range_min, min(x, range_max))

    def __next__(self):
        if self.random_character_spacing:
            raw_spacing = random.gauss(10, 40)
            self.character_spacing = round(self._clamp(raw_spacing, 0, 50))
        return self.next()


def generate_font_data(
    output_dir: PurePath,
    num_samples: int,
    font_path: PurePath,
    random_character_spacing: bool,
    backgrounds_path: Optional[PurePath],
    regex_template: Optional[str],
):
    # TODO: we have to generate the samples from the same strings.
    generator = (
        GeneratorFromRegex(
            count=num_samples,
            regex_template=regex_template,
            random_character_spacing=random_character_spacing,
            fonts=[str(font_path)],
            # This size maps to the height of the images.
            size=105,
            # Background == 1 means plain white and is the default
            # if no background images are provided.
            background_type=1 if not backgrounds_path else None,
            image_dir=backgrounds_path if backgrounds_path else None,
            fit=True,
        )
        if regex_template
        else GeneratorFromRandom(
            # Length works a bit oddly as it doesn't seem to map to
            # the length of the desired string.
            length=2,
            fonts=[str(font_path)],
            # This size maps to the height of the images.
            size=105,
            # Background == 1 means plain white and is the default
            # if no background images are provided.
            background_type=1 if not backgrounds_path else None,
            image_dir=backgrounds_path if backgrounds_path else None,
            fit=True,
        )
    )

    # Store the data in a subdirectory named as the font file.
    font_dir_name = font_path.stem
    output_dir = PurePath(output_dir, font_dir_name)

    # Make sure the output dir exists.
    Path(output_dir).mkdir(parents=True, exist_ok=True)

    labels_file = PurePath(output_dir, "labels.csv")
    with open(labels_file, "w", encoding="utf8") as csv:
        csv.write("filename,words\n")

        for i, (img, label) in enumerate(generator):
            filename = f"{i}.png"
            img.save(str(PurePath(output_dir, filename)))
            csv.write(f"{filename},{label}\n")

            if i == num_samples:
                break


def get_parser():
    parser = argparse.ArgumentParser(description="Fontina dataset generator")
    parser.add_argument(
        "-c",
        "--config",
        metavar="CONFIG-FILE",
        help="path to the configuration file",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--outdir",
        default="out-data",
        metavar="OUTPUT-DIR",
        help="path to the directory that will contain the outputs",
    )
    return parser


def main():
    args = get_parser().parse_args()

    config = load_config(args.config)

    fonts_config = config["fonts"]
    for f in fonts_config["classes"]:
        print(f"Generating font data for {f['path']}")
        generate_font_data(
            output_dir=PurePath(args.outdir),
            num_samples=fonts_config["samples_per_font"],
            font_path=PurePath(f["path"]),
            backgrounds_path=PurePath(fonts_config["backgrounds_path"])
            if "backgrounds_path" in fonts_config
            else None,
            random_character_spacing=fonts_config["random_character_spacing"],
            regex_template=fonts_config["regex_template"],
        )


if __name__ == "__main__":
    main()



import numpy as np
import torch

from torch.utils.data import Dataset


class AugmentedDataset(Dataset):
    def __init__(self, dataset, num_classes, transform=None):
        self.dataset = dataset
        self.num_classes = num_classes
        # Transforms are albumentations
        self.transform = transform

    def __getitem__(self, index):
        raw_image = self.dataset[index][0]
        x = (
            self.transform(image=np.asarray(raw_image))["image"]
            if self.transform
            else raw_image
        )
        return x, torch.as_tensor(self.dataset[index][1], dtype=torch.long)

    def __len__(self):
        return len(self.dataset)



import argparse
import numpy as np
import PIL
import torch
import torch.nn.functional as F

from fontina.augmentation_utils import (
    get_random_square_patch,
    get_test_augmentations,
    resize_fixed_height,
)

from fontina.models.deepfont import DeepFont, DeepFontAutoencoder
from fontina.models.lightning_wrappers import DeepFontWrapper


def get_parser():
    parser = argparse.ArgumentParser(description="Fontina predict")
    parser.add_argument(
        "-w",
        "--weights",
        metavar="WEIGHTS-FILE",
        help="path to the weights/checkpoint file",
        required=True,
    )
    parser.add_argument(
        "-n",
        "--num-classes",
        type=int,
        help="the number of classes",
        required=True,
    )
    parser.add_argument(
        "-i",
        "--input",
        metavar="INPUT-FILE",
        help="path to the input image file",
        required=True,
    )
    return parser


def predict(model: DeepFontWrapper, img) -> torch.Tensor:
    all_soft_preds = []
    for _ in range(3):
        enhancement_pipeline = get_test_augmentations(
            squeeze_ratio=np.random.uniform(low=1.5, high=3.5)
        )
        enhanced_img = enhancement_pipeline(image=np.asarray(img))["image"]

        patch_sampler = get_random_square_patch()
        patches = [patch_sampler(image=enhanced_img)["image"] for _ in range(5)]
        inputs = torch.tensor(np.asarray(patches))

        preds = model(inputs.cuda())
        soft_preds = F.softmax(preds, dim=1)
        all_soft_preds.append(soft_preds)

    return torch.cat(all_soft_preds).mean(0)


def main():
    args = get_parser().parse_args()

    model = DeepFontWrapper.load_from_checkpoint(
        args.weights,
        model=DeepFont(autoencoder=DeepFontAutoencoder(), num_classes=args.num_classes),
        num_classes=args.num_classes,
    )

    raw_img = PIL.Image.open(args.input).convert("L")
    img = resize_fixed_height(np.asarray(raw_img))

    predicted_class = predict(model, img)

    print(f"Mean softmax vector:\n{predicted_class}")
    print(f"Final Predicted label: {predicted_class.argmax()}")
    print(f"Sorted labels: {predicted_class.argsort(dim=0, descending=True)}")


if __name__ == "__main__":
    main()



import torch
import torch.nn as nn
import torch.nn.functional as F


class DeepFontAutoencoder(nn.Module):
    def __init__(self):
        super(DeepFontAutoencoder, self).__init__()

        # Encoder
        self.encoder = nn.Sequential(
            nn.Conv2d(
                in_channels=1, out_channels=64, kernel_size=12, stride=2, padding=1
            ),
            nn.ReLU(),
            nn.BatchNorm2d(64),
            nn.MaxPool2d(2),
            nn.Conv2d(
                in_channels=64, out_channels=128, kernel_size=3, stride=1, padding=1
            ),
            nn.ReLU(),
            nn.BatchNorm2d(128),
            nn.MaxPool2d(2),
        )

        # Decoder
        self.decoder = nn.Sequential(
            # See https://stackoverflow.com/a/58207809/261698
            nn.Upsample(scale_factor=2, mode="bilinear"),
            nn.ConvTranspose2d(
                in_channels=128, out_channels=64, kernel_size=3, stride=1, padding=1
            ),
            nn.ReLU(),
            nn.Upsample(scale_factor=2, mode="bilinear"),
            nn.ConvTranspose2d(
                in_channels=64,
                out_channels=1,
                kernel_size=12,
                stride=2,
                padding=1,
                output_padding=1,
            ),
            nn.Sigmoid(),
        )

    def forward(self, x):
        encoded = self.encoder(x)
        decoded = self.decoder(encoded)
        return decoded


class DeepFont(nn.Module):
    def __init__(self, autoencoder: DeepFontAutoencoder, num_classes: int):
        super(DeepFont, self).__init__()

        # Cross-domain subnetwork layers (Cu)
        # This is coming from the encoder.
        self.ae_encoder = autoencoder.encoder
        # Make sure we don't train the autoencoder again.
        for param in self.ae_encoder.parameters():
            param.requires_grad = False

        # Domain-specific layers (Cs)
        self.conv5 = nn.Conv2d(
            in_channels=128, out_channels=256, kernel_size=3, padding=1
        )
        self.conv6 = nn.Conv2d(
            in_channels=256, out_channels=256, kernel_size=3, padding=1
        )
        self.conv7 = nn.Conv2d(
            in_channels=256, out_channels=256, kernel_size=3, padding=1
        )

        self.flatten = nn.Flatten()

        hidden_units = 4096
        self.fc1 = nn.Linear(256 * 12 * 12, hidden_units)
        self.drop1 = nn.Dropout(0.5)
        self.fc2 = nn.Linear(hidden_units, hidden_units)
        self.drop2 = nn.Dropout(0.5)
        self.fc3 = nn.Linear(hidden_units, num_classes)

    def forward(self, x):
        # Cu Layers
        with torch.no_grad():
            out = self.ae_encoder(x)

        # Cs Layers
        out = F.relu(self.conv5(out))
        out = F.relu(self.conv6(out))
        out = F.relu(self.conv7(out))

        out = self.flatten(out)

        out = F.relu(self.fc1(out))
        out = self.drop1(out)
        out = F.relu(self.fc2(out))
        out = self.drop2(out)
        out = self.fc3(out)

        return out



import fontina.augmentation_utils as au
import numpy as np


def test_resize_fixed_height():
    test_img = np.ones((70, 100), dtype=np.uint8)
    resized = au.resize_fixed_height(test_img, new_height=105)

    # Check for the expected height.
    assert resized.shape[0] == 105

    # Check that the ratio of the source image is kept.
    assert np.isclose(
        test_img.shape[0] / test_img.shape[1], resized.shape[0] / resized.shape[1]
    )


def test_split_patches():
    test_img = np.ones((70, 100), dtype=np.uint8)
    patches = au.split_patches_np(test_img, 50, False)

    # Check that we have the expected number of patches
    # of the expected size.
    assert len(patches) == 2
    assert all([p.shape[1] == 50 for p in patches])


def test_split_patches_partial():
    test_img = np.ones((70, 90), dtype=np.uint8)
    patches = au.split_patches_np(test_img, 50, False)

    # Even if the image width is shorter than 2 patches,
    # we still expect two of them: the split function will
    # fill part of the second one with full-white.
    assert len(patches) == 2
    assert all([p.shape[1] == 50 for p in patches])

    last_patch = patches[1]
    assert np.all(last_patch[:, 40:] == np.ones((70, 10), dtype=np.uint8) * 255)


def test_split_patches_drop():
    test_img = np.ones((70, 90), dtype=np.uint8)
    patches = au.split_patches_np(test_img, 50, True)

    # The image is too short for two patches and we explicitly
    # ask the function to drop the last partial patch.
    assert len(patches) == 1
    assert patches[0].shape[1] == 50


def test_pick_random_patch_constrained():
    # Craft an image with two 105x105 patches: one full black, one full white.
    test_img = np.ones((105, 110), dtype=np.uint8) * 255
    test_img[:, :105] = 0

    aug = au.PickRandomPatch(constrained_patches=True, always_apply=True)
    result = aug(image=test_img)["image"]

    assert np.all(np.equal(result.shape, [105, 105]))

    # It's either a full black or a full white patch, as we're constraining
    # the random choice into non-random patches.
    assert np.all(result == 255) or np.all(result == 0)


def test_pick_random_small_width():
    # Craft an image with one patch: its width being less than 105.
    test_img = np.ones((105, 80), dtype=np.uint8) * 255

    aug = au.PickRandomPatch(constrained_patches=True, always_apply=True)
    result = aug(image=test_img)["image"]

    # The returned patch must still be 105x105.
    assert np.all(np.equal(result.shape, [105, 105]))


def test_pick_random_exact_width():
    # Craft an image with one patch: its width being exactly 105.
    test_img = np.ones((105, 105), dtype=np.uint8) * 255

    aug = au.PickRandomPatch(constrained_patches=True, always_apply=True)
    result = aug(image=test_img)["image"]

    # The returned patch must still be 105x105.
    assert np.all(np.equal(result.shape, [105, 105]))



import pytorch_lightning as L
import torch
import torch.nn.functional as F

from .deepfont import DeepFont, DeepFontAutoencoder
from torchmetrics import Accuracy
from pytorch_lightning.utilities.grads import grad_norm


class DeepFontAutoencoderWrapper(L.LightningModule):
    def __init__(self):
        super().__init__()
        # Saving hyperparameters of autoencoder
        self.save_hyperparameters()
        self.autoencoder = DeepFontAutoencoder()
        # Example input array needed for visualizing the graph of the network
        self.example_input_array = torch.zeros(2, 1, 105, 105)

    def forward(self, x):
        return self.autoencoder(x)

    def _get_reconstruction_loss(self, batch):
        x, _ = batch
        x_hat = self.forward(x)
        return F.mse_loss(x, x_hat)

    def training_step(self, batch, batch_idx):
        loss = self._get_reconstruction_loss(batch)
        self.log("train_loss", loss, prog_bar=True)
        return loss

    def validation_step(self, batch, batch_idx):
        loss = self._get_reconstruction_loss(batch)
        self.log("val_loss", loss, prog_bar=True)

    def test_step(self, batch, batch_idx):
        loss = self._get_reconstruction_loss(batch)
        self.log("test_loss", loss, prog_bar=True)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
        # Using a scheduler is optional but can be helpful.
        # The scheduler reduces the LR if the validation
        # performance hasn't improved for the last N epochs
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            optimizer, mode="min", factor=0.2, patience=20, min_lr=5e-5
        )
        return {"optimizer": optimizer, "lr_scheduler": scheduler, "monitor": "val_loss"}


class DeepFontWrapper(L.LightningModule):
    def __init__(self, model: DeepFont, num_classes: int, learning_rate: float = 0.01):
        super().__init__()

        self.learning_rate = learning_rate
        self.model = model
        self.accuracy = Accuracy(task="multiclass", num_classes=num_classes, top_k=1)
        self.eval_loss: list[torch.Tensor] = []
        self.eval_accuracy: list[torch.Tensor] = []
        self.test_accuracy: list[torch.Tensor] = []
        self.example_input_array = torch.zeros(2, 1, 105, 105)

    def forward(self, x):
        return self.model(x)

    def training_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self(x)
        loss = F.cross_entropy(y_hat, y)

        train_acc = self.accuracy(y_hat.argmax(1), y)

        self.log("train_acc", train_acc, prog_bar=True)
        self.log("train_loss", loss, prog_bar=True)
        return loss

    def validation_step(self, batch, batch_idx):
        loss, acc = self._shared_eval(batch)
        self.log("val_loss", loss, prog_bar=True)
        self.log("val_accuracy", acc, prog_bar=True)
        self.eval_loss.append(loss)
        self.eval_accuracy.append(acc)
        return {"val_loss": loss, "val_accuracy": acc}

    def test_step(self, batch, batch_idx):
        loss, acc = self._shared_eval(batch)
        self.log_dict({"test_loss": loss, "test_acc": acc})

    def _shared_eval(self, batch):
        x, y = batch
        y_hat = self(x)
        loss = F.cross_entropy(y_hat, y)
        acc = self.accuracy(y_hat.argmax(1), y)
        return loss, acc

    def on_before_optimizer_step(self, optimizer):
        # Compute the 2-norm for each layer
        # If using mixed precision, the gradients are already unscaled here
        norms = grad_norm(self.model, norm_type=2)
        self.log_dict(norms)

    def configure_optimizers(self):
        params = filter(lambda p: p.requires_grad, self.parameters())
        optimizer = torch.optim.SGD(
            params, lr=self.learning_rate, momentum=0.9, weight_decay=0.0005
        )
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            optimizer, "min", patience=5
        )
        return {
            "optimizer": optimizer,
            "lr_scheduler": {
                "scheduler": scheduler,
                "monitor": "val_loss",
            },
        }



import yaml


def load_config(conf_path: str) -> dict:
    with open(conf_path, "r", encoding="utf8") as c:
        try:
            return yaml.safe_load(c)
        except yaml.YAMLError as e:
            # TODO: provide better error handling
            print(e)
            return {}



import cv2
import io
import numpy as np
import torch

from torch.utils.data import Dataset


def read_label(label_file: str):
    with open(label_file, "rb") as f:
        # We need to cast to `int32` as PyTorch 2.0.1 does not yet
        # support collating `uin16`, `uint32` in data loaders. See
        # https://github.com/pytorch/pytorch/issues/58734
        return np.frombuffer(f.read(), np.uint32, -1).astype("int32")


class AdobeVFRDataset(Dataset):
    def __init__(self, bcf_path: str, dataset_type: str, transform=None):
        self.bcf_filename = f"{bcf_path}/{dataset_type}.bcf"
        self.labels = read_label(f"{bcf_path}/{dataset_type}.label")
        self.num_labels = len(np.unique(self.labels))
        # Transforms are albumentations
        self.transform = transform

        self._load_bcf_meta()

    def __getitem__(self, index):
        binary_image = self._get_bcf_entry_by_index(index)
        image_as_array = np.asarray(
            bytearray(io.BytesIO(binary_image).read()), dtype=np.uint8
        )
        cv2image = cv2.imdecode(image_as_array, cv2.IMREAD_GRAYSCALE)
        raw_image = np.array(cv2image, dtype="uint8")
        x = self.transform(image=raw_image)["image"] if self.transform else raw_image
        # We need to cast to `torch.long` to prevent errors such as
        # "nll_loss_forward_reduce_cuda_kernel_2d_index" not implemented for 'Int'.
        return x, torch.as_tensor(self.labels[index], dtype=torch.long)

    def __len__(self):
        return len(self._bcf_offsets) - 1

    def _load_bcf_meta(self):
        with open(self.bcf_filename, "rb") as f:
            # Read the size of the section with files?
            size = np.frombuffer(f.read(8), dtype=np.uint64)[0]
            # Must cast for some reason, otherwise this yields a float64.
            file_sizes = np.frombuffer(f.read(int(8 * size)), dtype=np.uint64)
            self._bcf_offsets = np.append(np.uint64(0), np.add.accumulate(file_sizes))

    def _get_bcf_entry_by_index(self, i):
        with open(self.bcf_filename, "rb") as f:
            f.seek((len(self._bcf_offsets) * 8 + self._bcf_offsets[i]).astype("uint64"))
            return f.read(self._bcf_offsets[i + 1] - self._bcf_offsets[i])



import torch
import torchvision

from pytorch_lightning import Callback


# This is basically a copy of
# https://lightning.ai/docs/pytorch/stable/notebooks/course_UvA-DL/08-deep-autoencoders.html#Training-the-model
class GenerateCallback(Callback):
    def __init__(self, input_imgs, every_n_epochs=1):
        super().__init__()
        self.input_imgs = input_imgs  # Images to reconstruct during training
        # Only save those images every N epochs (otherwise tensorboard gets quite large)
        self.every_n_epochs = every_n_epochs

    def on_train_epoch_end(self, trainer, pl_module):
        if trainer.current_epoch % self.every_n_epochs == 0:
            # Reconstruct images
            input_imgs = self.input_imgs.to(pl_module.device)
            with torch.no_grad():
                pl_module.eval()
                reconst_imgs = pl_module(input_imgs)
                pl_module.train()
            # Plot and add to tensorboard
            imgs = torch.stack([input_imgs, reconst_imgs], dim=1).flatten(0, 1)
            grid = torchvision.utils.make_grid(
                imgs, nrow=2, normalize=True, range=(-1, 1)
            )
            trainer.logger.experiment.add_image(
                "Reconstructions", grid, global_step=trainer.global_step
            )
